/*
    $Id: scripting.c,v 1.11 2001/12/21 10:28:13 sensei Exp $

    Copyright (C) 2000-2001  Marko Zivanovic
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
     
#ifndef lint
static char vcid[] __attribute__ ((__unused__)) = "$Id: scripting.c,v 1.11 2001/12/21 10:28:13 sensei Exp $";
#endif /* lint */

#include "scripting.h"

/* script interpreter */
Tcl_Interp *S_int;

/* script mutex device */
pthread_mutex_t S_mutex;

int _dump_eth = 0, _dump_ip = 0, _dump_icmp = 0, _dump_tcp = 0, _dump_udp = 0;

/*  ************************************************************ */
/*  initialize global scripting engine                           */
/*                                                               */
/*  return OK                                                    */
/*  ************************************************************ */
int S_init_scripting ()
{
  int ret;

  S_int = Tcl_CreateInterp();
  
  ret = pthread_mutex_init (&S_mutex, NULL);
  if (ret != 0) {
    logmsg (LOG_CRIT, "error creating mutex: %s", strerror (errno));
    exit (2);
  }

  return TCL_OK;
}

/*  ************************************************************ */
/*  load script file into specified interpreter                  */
/*                                                               */
/*  parameters:                                                  */
/*     script_file -  script file to load into script interpreter*/
/*                                                               */
/*  return:                                                      */
/*     int       -  return code                                  */
/*  ************************************************************ */
int S_load_script (Tcl_Interp* interp, char* script_file)
{
  int ret;
  
  ret = Tcl_EvalFile (interp, script_file);
  if (ret != TCL_OK) {
    logmsg (LOG_CRIT, "script error: %s", S_getErrorInfo (interp));
    exit (2);
  }
  
  return TCL_OK;
}  

/*  ************************************************************ */
/*  call script function from specified interpreter              */
/*  ************************************************************ */
int S_call (Tcl_Interp* interp, char *function)
{
  int ret;

  ret = Tcl_Eval (interp, function);
  if (ret != TCL_OK) {
    logmsg (LOG_CRIT, "error in %s: %s", function, S_getErrorInfo (interp));
    exit (2);
  }
  
  return ret;
}

/*  ************************************************************ */
/*  call on_dump_packet event handler from script                */
/*                                                               */
/*  no parameters, no return code                                */
/*  ************************************************************ */
int S_on_dump_packet (Tcl_Interp* interp, eth_ip_frame* eth_pkt)
{
  int ret;
 
  S_load_all_headers (interp, eth_pkt);
 
  ret = Tcl_Eval (interp, "sniffer::on_dump_packet");
  if (ret != TCL_OK) {
    logmsg (LOG_CRIT, "error in on_dump_packet: %s", S_getErrorInfo(interp));
    exit (2);
  }
 
  return atoi (S_int->result);
}


/*  ************************************************************ */
/*  call on_receive_packet event handler from script             */
/*                                                               */
/*  no parameters, no return code                                */
/*  ************************************************************ */
char* S_on_receive_packet (Tcl_Interp* interp, eth_ip_frame* eth_pkt)
{
  int ret;
  
  S_load_all_headers (interp, eth_pkt);
  
  ret = Tcl_Eval (interp, "on_receive_packet");
  if (ret != TCL_OK) {
    logmsg (LOG_CRIT, "error in on_receive_packet: %s", S_getErrorInfo(interp));
    exit (2);
  }
  
  return interp->result;
}

/*  ************************************************************ */
/*  set script variable to string value at the current level of  */
/*  execution (usually global)                                   */
/*                                                               */
/*  parameters:                                                  */
/*     varname   -   name of the variable to set                 */
/*     value     -   string value to assign to the variable      */
/*                                                               */
/*  no return                                                    */
/*  ************************************************************ */
int S_set_str_var (Tcl_Interp* interp, char *varname, char *value)
{
  char tmp[1024*64];
  int ret;

  snprintf (tmp, 64*1024, "set %s {%s}", varname, value);
  ret = Tcl_Eval (interp, tmp);
  if (ret != TCL_OK)    
    return 1;

  return 0;
}

/*  ************************************************************ */
/*  set script variable to integer value at the current level of */
/*  execution (usually global)                                   */
/*                                                               */
/*  parameters:                                                  */
/*     varname   -   name of the variable to set                 */
/*     value     -   integer value to assign to the variable     */
/*                                                               */
/*  no return                                                    */
/*  ************************************************************ */
int S_set_int_var (Tcl_Interp* interp, char *varname, int value)
{
  char tmp[1024];
  int ret;

  snprintf (tmp, 1024, "set %s %d", varname, value);
  ret = Tcl_Eval (interp, tmp);
  if (ret != TCL_OK)
    return 1;

  return 0;
}

/*  ************************************************************ */
/*  retreive string value of the script variable at the current  */
/*  level of execution                                           */
/*                                                               */
/*  parameters:                                                  */
/*     name   -   name of variable to read                       */
/*                                                               */
/*  return:                                                      */
/*     char*  -   string value of the variable                   */
/*  ************************************************************ */
char* S_get_str_var (Tcl_Interp* interp, char* name)
{
  char *val;

  val = Tcl_GetVar (interp, name, 0);
  if (val == NULL) {
    logmsg (LOG_CRIT, "error reading variable: %s", S_getErrorInfo(interp));
    exit(2);
  }

  return val;
}

/*  ************************************************************ */
/*  retreive integer value of the script variable at the current */
/*  level of execution                                           */
/*                                                               */
/*  parameters:                                                  */
/*     name   -   name of variable to read                       */
/*                                                               */
/*  return:                                                      */
/*     int    -   integer value of the variable                  */
/*  ************************************************************ */
int S_get_int_var (Tcl_Interp* interp, char* name)
{
  char *val;

  val = Tcl_GetVar (interp, name, 0);
  if (val == NULL) {
    logmsg (LOG_CRIT, "error reading variable %s: %s", name, S_getErrorInfo(interp));
    exit(2);
  }

  return atoi (val);
}

int S_hex_dump (char* name, char* source, int len)
{
  /* 
     FIXME: this solution is way too ugly; replace it with dynamic
     (re)allocation later 
  */
 
#define MAX_CODED_LEN 1024 * 64

  char tmp[5], out[MAX_CODED_LEN];
  int x;
  
  out[0] = 0;
  
  for (x = 0; x < len; x++) {
    sprintf (tmp, "%2.2hhX ", (unsigned char) source[x]);
    strcat (out, tmp);
  }
  
  //  S_set_str_var (name, (char*) &out);
  
  return 0;
}

int S_ascii_dump (char* name, char* source, int len)
{
  /* 
     FIXME: this solution is way too ugly; replace it with dynamic
     (re)allocation later 
  */
 
#define MAX_CODED_LEN 1024 * 64

  char tmp[5], out[MAX_CODED_LEN];
  int x;
  
  out[0] = 0;
    
  for (x = 0; x < len; x++) {
    if (isprint ((unsigned char) source[x])) {
      sprintf (tmp, "%c", (unsigned char) source[x]);
    } else {
      sprintf (tmp, ".");
    }
    strcat (out, tmp);
  }
  
  //  S_set_str_var (name, (char*) &out);
  
  return 0;
}

/*  ************************************************************ */
/*  decode packet headers and load script variables with         */
/*  appropriate values                                           */
/*                                                               */
/*  parameters:                                                  */
/*     eth_ip   -   ethernet frame to be decoded                 */
/*                                                               */
/*  no return                                                    */
/*  ************************************************************ */
int S_load_all_headers (Tcl_Interp* interp, eth_ip_frame* eth_ip)
{
  struct iphdr* ip;

  /* check whether ethernet frame is carrying IP packet */
  if (ntohs (eth_ip->eth.h_proto) != ETH_P_IP) {
    _dump_eth = 1;
  } else {
    ip = (struct iphdr*) (((unsigned long) &eth_ip->ip) - 2);
    if (ip->version != 4) {
      _dump_eth = 1;
    } else {
      
      switch (ip->protocol) {	      
      case 1:   /* ICMP */
	_dump_eth = _dump_ip = _dump_icmp = 1; break;
      case 6:   /* TCP */
	_dump_eth = _dump_ip = _dump_tcp = 1; break;
      case 17:   /* UDP */
	_dump_eth = _dump_ip = _dump_udp = 1; break;
      default:   /* all other IP protocols */
	_dump_eth = _dump_ip = 1; break;
	
      } /* switch ip protocols */
    } /* if ip->version == 4 */
  } /* if ethernet protocol == ip */	
  
  /* load script interpreter with apropriate headers broken-down to fields */
  S_load_packet_info (interp, (eth_ip_frame*) eth_ip);
  if (_dump_eth == 1) S_load_eth_hdr (interp, (eth_ip_frame*) eth_ip);
  if (_dump_ip == 1) S_load_ip_hdr (interp, (eth_ip_frame*) eth_ip);
  if (_dump_icmp == 1) S_load_icmp_hdr (interp, (eth_ip_icmp_frame*) eth_ip);
  if (_dump_tcp == 1) S_load_tcp_hdr (interp, (eth_ip_tcp_frame*) eth_ip);
  if (_dump_udp == 1) S_load_udp_hdr (interp, (eth_ip_udp_frame*) eth_ip);

  return 0;
}

/*  ************************************************************ */
/*  load general packet info into script interpreter             */
/*                                                               */
/*  parameters:                                                  */
/*     eth_ip_frame*   -   packet whose info to insert into      */
/*                         script interpreter                    */
/*                                                               */
/*  no return code                                               */
/*  ************************************************************ */
int S_load_packet_info (Tcl_Interp* interp, eth_ip_frame* eth_pkt)
{
  int len;

  /* total length of the packet headers and payload */
  len = eth_pkt->length - sizeof (eth_pkt->time) - sizeof (eth_pkt->length);
  S_set_int_var (interp, "packet(length)", len);
  S_set_str_var (interp, "packet(type)", "ethernet");
  S_set_int_var (interp, "packet(time)", eth_pkt->time);
  //  S_hex_dump (interp, "packet(raw)", (char*) &eth_pkt->eth, len);

  return 0;
}

/*  ************************************************************ */
/*  load ethernet header fields from the specified packet into   */
/*  script interpreter                                           */
/*                                                               */
/*  parameters:                                                  */
/*    eth_ip_frame*   -   packet containing ethernet header      */
/*                                                               */
/*  no return code                                               */
/*  ************************************************************ */
int S_load_eth_hdr (Tcl_Interp* interp, eth_ip_frame *eth_pkt)
{
  struct ethhdr *eth;

  eth = (struct ethhdr*) (((unsigned long) &eth_pkt->eth));

  /*    load ethernet header fields into script interpreter */
  S_set_str_var (interp, "eth(saddr)", eth_ntoa (eth->h_source));
  S_set_str_var (interp, "eth(daddr)", eth_ntoa (eth->h_dest));
  S_set_int_var (interp, "eth(proto)", ntohs (eth->h_proto));

  return 0;
}
  
/*  ************************************************************ */
/*  load IP header fields from the specified packet into script  */
/*  interpreter                                                  */
/*                                                               */
/*  parameters:                                                  */
/*    eth_ip_frame*   -   packet containing IP header            */
/*                                                               */
/*  no return code                                               */
/*  ************************************************************ */
int S_load_ip_hdr (Tcl_Interp* interp, eth_ip_frame *eth_pkt)
{
  struct ethhdr *eth;
  struct iphdr *ip;
  struct in_addr in;
  int len;

  eth = (struct ethhdr*) (((unsigned long) &eth_pkt->eth));
  ip = (struct iphdr*) (((unsigned long) &eth_pkt->ip) - 2);

/*    load ip header fields into script interpreter */
  S_set_int_var (interp, "ip(version)", ip->version);
  S_set_int_var (interp, "ip(ihl)", ip->ihl);
  S_set_int_var (interp, "ip(tos)", ip->tos);
  S_set_int_var (interp, "ip(tot_len)", ntohs (ip->tot_len));
  S_set_int_var (interp, "ip(id)", ntohs (ip->id));

/*   TODO:
     - add support for IP header flags (in frag_off)
*/

  S_set_int_var (interp, "ip(frag_off)", ntohs (ip->frag_off));
  S_set_int_var (interp, "ip(ttl)", ip->ttl);
  S_set_int_var (interp, "ip(protocol)", ip->protocol);
  S_set_int_var (interp, "ip(check)", ip->check);
  memcpy (&in.s_addr, &ip->saddr, sizeof (&ip->saddr));
  S_set_str_var (interp, "ip(saddr)", inet_ntoa (in));
  memcpy (&in.s_addr, &ip->daddr, sizeof (&ip->daddr));
  S_set_str_var (interp, "ip(daddr)", inet_ntoa (in));

  len = htons (ip->tot_len) - sizeof (eth_pkt->ip);
  //  S_hex_dump ("ip(data_hex)", (char*) &eth_pkt->data - 2, len);
  //  S_ascii_dump ("ip(data_ascii)", (char*) &eth_pkt->data - 2, len);
  S_set_int_var (interp, "ip(data_length)", len);

  return 0;
}

/*  ************************************************************ */
/*  load icmp header fields from the specified packet into       */
/*  script interpreter                                           */
/*                                                               */
/*  parameters:                                                  */
/*    eth_ip_icmp_frame*   -   packet containing icmp header     */
/*                                                               */
/*  no return code                                               */
/*  ************************************************************ */
int S_load_icmp_hdr (Tcl_Interp* interp, eth_ip_icmp_frame *eth_pkt)
{
  struct ethhdr *eth;
  struct iphdr *ip;
  struct icmphdr *icmp;
  int len;

  eth = (struct ethhdr*) (((unsigned long) &eth_pkt->eth));
  ip = (struct iphdr*) (((unsigned long) &eth_pkt->ip - 2));
  icmp = (struct icmphdr*) (((unsigned long) &eth_pkt->icmp - 2));

  S_set_int_var (interp, "icmp(type)", icmp->type);
  S_set_str_var (interp, "icmp(type_str)", icmp_type (icmp->type));
  S_set_int_var (interp, "icmp(code)", icmp->code);
  S_set_str_var (interp, "icmp(code_str)", icmp_code (icmp->type, icmp->code));
  S_set_int_var (interp, "icmp(checksum)", ntohs (icmp->checksum));
  S_set_int_var (interp, "icmp(echo_id)", ntohs (icmp->un.echo.id));
  S_set_int_var (interp, "icmp(echo_sequence)", ntohs (icmp->un.echo.sequence));
  S_set_int_var (interp, "icmp(gateway)", ntohl (icmp->un.gateway));
  S_set_int_var (interp, "icmp(frag_unused)", ntohs (icmp->un.frag.__unused));
  S_set_int_var (interp, "icmp(frag_mtu)", ntohs (icmp->un.frag.mtu));

  len = htons (ip->tot_len) - sizeof (eth_pkt->ip) - sizeof (eth_pkt->icmp);
  //  S_hex_dump ("icmp(data_hex)", (char*) &eth_pkt->data - 2, len);
  //  S_ascii_dump ("icmp(data_ascii)", (char*) &eth_pkt->data - 2, len);
  S_set_int_var (interp, "icmp(data_length)", len);

  return 0;
}

/*  ************************************************************ */
/*  load TCP header fields from the specified packet into script */
/*  interpreter                                                  */
/*                                                               */
/*  parameters:                                                  */
/*    eth_ip_tcp_frame*   -   packet containing tcp header       */
/*                                                               */
/*  no return code                                               */
/*  ************************************************************ */
int S_load_tcp_hdr (Tcl_Interp* interp, eth_ip_tcp_frame *eth_pkt)
{
  struct ethhdr *eth;
  struct iphdr *ip;
  struct tcphdr *tcp;
  int len;

  eth = (struct ethhdr*) (((unsigned long) &eth_pkt->eth));
  ip = (struct iphdr*) (((unsigned long) &eth_pkt->ip - 2));
  tcp = (struct tcphdr*) (((unsigned long) &eth_pkt->tcp - 2));

  S_set_int_var (interp, "tcp(source)", ntohs (tcp->source));
  S_set_int_var (interp, "tcp(dest)", ntohs (tcp->dest));
  S_set_int_var (interp, "tcp(seq)", ntohl (tcp->seq));
  S_set_int_var (interp, "tcp(ack_seq)", ntohl (tcp->ack_seq));
  S_set_int_var (interp, "tcp(doff)", tcp->doff);
  S_set_int_var (interp, "tcp(fin)", tcp->fin);
  S_set_int_var (interp, "tcp(syn)", tcp->syn);
  S_set_int_var (interp, "tcp(rst)", tcp->rst);
  S_set_int_var (interp, "tcp(psh)", tcp->psh);
  S_set_int_var (interp, "tcp(ack)", tcp->ack);
  S_set_int_var (interp, "tcp(urg)", tcp->urg);
  S_set_int_var (interp, "tcp(window)", ntohs (tcp->window));
  S_set_int_var (interp, "tcp(check)", ntohs (tcp->check));
  S_set_int_var (interp, "tcp(urg_ptr)", ntohs (tcp->urg_ptr));

  len = htons (ip->tot_len) - sizeof (eth_pkt->ip) - sizeof (eth_pkt->tcp);
  //  S_hex_dump ("tcp(data_hex)", (char*) &eth_pkt->data - 2, len);
  //  S_ascii_dump ("tcp(data_ascii)", (char*) &eth_pkt->data - 2, len);
  S_set_int_var (interp, "tcp(data_length)", len);

  return 0;
}

/*  ************************************************************ */
/*  load udp header fields from the specified packet into script */
/*  interpreter                                                  */
/*                                                               */
/*  parameters:                                                  */
/*    eth_ip_udp_frame*   -   packet containing udp header       */
/*                                                               */
/*  no return code                                               */
/*  ************************************************************ */
int S_load_udp_hdr (Tcl_Interp* interp, eth_ip_udp_frame *eth_pkt)
{
  struct ethhdr *eth;
  struct iphdr *ip;
  struct udphdr *udp;
  int len;

  eth = (struct ethhdr*) (((unsigned long) &eth_pkt->eth));
  ip = (struct iphdr*) (((unsigned long) &eth_pkt->ip - 2));
  udp = (struct udphdr*) (((unsigned long) &eth_pkt->udp - 2));

  S_set_int_var (interp, "udp(source)", ntohs (udp->source));
  S_set_int_var (interp, "udp(dest)", ntohs (udp->dest));
  S_set_int_var (interp, "udp(len)", ntohs (udp->len));
  S_set_int_var (interp, "udp(check)", ntohs (udp->check));

  len = htons (ip->tot_len) - sizeof (eth_pkt->ip) - sizeof (eth_pkt->udp);
  //  S_hex_dump ("udp(data_hex)", (char*) &eth_pkt->data - 2, len);
  //  S_ascii_dump ("udp(data_ascii)", (char*) &eth_pkt->data - 2, len);
  S_set_int_var (interp, "udp(data_length)", len);

  return 0;
}

/*  ************************************************************ */
/*  retreive extended error information from script interpreter  */
/*                                                               */
/*  no parameters                                                */
/*  return error info                                            */
/*  ************************************************************ */
char* S_getErrorInfo (Tcl_Interp* interp)
{
  char *ret;

  ret = Tcl_GetVar (interp, "errorInfo", TCL_GLOBAL_ONLY);
  /* TODO: add error checking */

  return ret;
}
  
/*  ************************************************************ */
/*  load custom commands into script interpreter                 */
/*                                                               */
/*  no parameters, no return code                                */
/*  ************************************************************ */
int S_load_cmds ()
{
  /*
    TODO:
    - add checks for return values throughout following functions
  */

  Tcl_CreateCommand (S_int, "promisc_on", &S_promiscOn, 0, NULL);
  Tcl_CreateCommand (S_int, "promisc_off", &S_promiscOff, 0, NULL);

  return 0;
}

/*  ************************************************************ */
/*  switch device we're listening to to promiscuous mode; this   */
/*  function is called whenever script interpreter encounters    */
/*  specified keyword ("promisc_on")                             */
/*                                                               */
/*  parameters:                                                  */
/*     ClientData   -   one word containing custom data,         */
/*                      currently unused                         */
/*     Tcl_Interp*  -   interpeter in wich we're executing       */
/*     argc         -   number of parameters passed to our       */
/*                      procedure from the script (we expect     */
/*                      zero)                                    */
/*     argv         -   list of parameters passed from script    */
/*                                                               */
/*  return TCL_OK or TCL_ERROR back to script interpreter        */
/*  ************************************************************ */
int S_promiscOn (ClientData cd, Tcl_Interp* S_int, int argc,
		 char **argv)
{
  char tmp[1024];
  
  /*  first parameter is the name of the function and we skip it */
  argc--;
  argv++;
  
  if (argc != 0) {
    snprintf (tmp, 1024, "invalid number of parameters (expect 0 got %d)", argc);
    strcpy (S_int->result, tmp);
    return TCL_ERROR;
  }
  
  promisc_on (sock, device);
  
  return TCL_OK;
}

/*  ************************************************************ */
/*  switch device we're listening to from promiscuous mode; this */
/*  function is called whenever script interpreter encounters    */
/*  specified keyword ("promisc_off")                            */
/*                                                               */
/*  parameters:                                                  */
/*     ClientData   -   one word containing custom data,         */
/*                      currently unused                         */
/*     Tcl_Interp*  -   interpeter in wich we're executing       */
/*     argc         -   number of parameters passed to our       */
/*                      procedure from the script (we expect     */
/*                      zero)                                    */
/*     argv         -   list of parameters passed from script    */
/*                                                               */
/*  return TCL_OK or TCL_ERROR back to script interpreter        */
/*  ************************************************************ */
int S_promiscOff (ClientData cd, Tcl_Interp* S_int, int argc,
		  char **argv)
{
  char tmp[1024];

  /*  first parameter is the name of the function and we skip it */
  argc--;
  argv++;

  if (argc != 0) {
    snprintf (tmp, 1024, "invalid number of parameters (expect 0 got %d)", argc);
    strcpy (S_int->result, tmp);
    return TCL_ERROR;
  }

  promisc_off (sock, device);

  return TCL_OK;
}
